use clichess;
use clichess::{Player, RecvPositionError, UserRole, EXIT_MSG};
use serde_json::Value;
use shakmaty::{fen, Chess, Color, Setup};
use std::collections::HashMap;
use std::fs;
use std::io;
use std::os::unix::net::{UnixListener, UnixStream};
use std::sync::mpsc::{channel, Receiver, TryRecvError};
use std::sync::{Arc, Mutex};

use std::thread;
use std::time;

struct Server {
    id: usize,
    chess_position: Arc<Mutex<Chess>>,
    players: Arc<Mutex<HashMap<usize, (Player, Arc<Mutex<String>>)>>>,
    others_serv_msg_buf: Arc<Mutex<String>>,
    client_message_recv: Receiver<String>,
}

fn main() {
    let chess = Arc::new(Mutex::new(Chess::default()));
    let players = Arc::new(Mutex::new(HashMap::new()));
    let mut counter = 0;
    fs::remove_file("/tmp/clichess.socket");

    let listener = UnixListener::bind("/tmp/clichess.socket").unwrap();

    // accept connections and process them, spawning a new thread for each one
    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                let client_message_recv = setup_client_message_recv(&stream).unwrap();
                let server = Server {
                    id: counter,
                    chess_position: chess.clone(),
                    players: players.clone(),
                    others_serv_msg_buf: Arc::new(Mutex::new(String::new())),
                    client_message_recv,
                };
                /* connection succeeded */
                thread::spawn(move || handle_player(stream, server));
                counter += 1;
            }
            Err(err) => {
                println!("Could not create a connection: {}", err);
                break;
            }
        }
    }
}

fn handle_player(mut stream: UnixStream, server: Server) {
    match initialize_client(&mut stream, &server) {
        Ok((player, player_turn)) => main_loop(&mut stream, &server, player, player_turn),
        Err(e) => println!("User id {} could not be initialized: {}", server.id, e),
    };
    player_disconnected(&server);
}

fn initialize_client(
    stream: &mut UnixStream,
    server: &Server,
) -> Result<(Player, Color), RecvPositionError> {
    //create player
    let player = create_player(server, stream)?;
    let player_turn: Color;
    //send current position to the player
    println!(
        "server {}, send current position to the player...",
        server.id
    );
    {
        let chess = server.chess_position.lock().unwrap();
        player_turn = chess.turn();
        clichess::write_to_stream(stream, fen::fen(&*chess)).unwrap();
    }
    println!("server {}, current position to the player sent", server.id);
    Ok((player, player_turn))
}

fn main_loop(stream: &mut UnixStream, server: &Server, player: Player, mut player_turn: Color) {
    loop {
        if clichess::is_player_turn(&player, player_turn) {
            //let go of the lock while waiting for user input.
            println!("server {}, waiting for player move..", server.id);
            let input;
            match server.client_message_recv.recv() {
                Ok(i) => input = i,
                Err(e) => {
                    println!("Error while getting user input: {}", e);
                    break;
                }
            };
            if input == EXIT_MSG {
                break;
            }
            {
                let mut chess = server.chess_position.lock().unwrap();
                let players = server.players.lock().unwrap();
                println!("User tried to play {}", input);
                match clichess::try_to_play_move(&chess, input) {
                    Ok(played_chess) => *chess = played_chess,
                    Err(e) => println!("Error: {}", e),
                };
                let chessfen = fen::fen(&*chess);
                for (id, (_, others_serv_msg_buf)) in players.iter() {
                    if server.id != *id {
                        others_serv_msg_buf.lock().unwrap().push_str(&chessfen);
                    }
                }
                clichess::write_to_stream(stream, chessfen).unwrap();
                player_turn = chess.turn();
            }
        } else {
            let position_as_fen_result = wait_for_opponent_move(server);
            if position_as_fen_result.is_err() {
                break;
            }
            let position_as_fen = position_as_fen_result.unwrap();
            println!("server id: {}, sending {}", server.id, position_as_fen);
            clichess::write_to_stream(stream, position_as_fen.clone()).unwrap();
            let chess = server.chess_position.lock().unwrap();
            player_turn = chess.turn();
        }
    }
}

fn wait_for_opponent_move(server: &Server) -> Result<String, RecvPositionError> {
    println!("server id: {}, waiting for move to send...", server.id);
    //wait: either we receive next position from other server threads, or we receive
    //"exit" from the client.
    let returned_result: Result<String, RecvPositionError>;
    loop {
        {
            let mut others_serv_msg_buf = server.others_serv_msg_buf.lock().unwrap();
            if !others_serv_msg_buf.is_empty() {
                returned_result = Ok(others_serv_msg_buf.clone());
                others_serv_msg_buf.clear();
                break;
            } else {
                match server.client_message_recv.try_recv() {
                    Ok(msg) => {
                        if msg == EXIT_MSG {
                            returned_result = Err(RecvPositionError::UserCanceledError);
                            break;
                        } else {
                            println!(
                                "Client sent message while it's not its turn, this is an error."
                            );
                            println!("Here is the message: {}", msg);
                            continue;
                        }
                    }
                    Err(TryRecvError::Disconnected) => println!("Error: client disconnected."),
                    Err(TryRecvError::Empty) => continue,
                }
            }
        }
        thread::sleep(time::Duration::from_millis(10));
    }
    returned_result
}

fn create_player(server: &Server, stream: &mut UnixStream) -> Result<Player, RecvPositionError> {
    println!("Creating player {}...", server.id);
    //get player name and pubkey
    let username_pubkey_json = server
        .client_message_recv
        .recv()
        .expect("Player closed connection.");
    let username_pubkey_value: Value = serde_json::from_str(&username_pubkey_json).unwrap();
    let username = username_pubkey_value["username"].to_string();
    let public_key = username_pubkey_value["pubkey"].to_string();
    println!("got username: {}", username);
    println!("got pubkey: {}", public_key);
    let role = receive_user_role(server, stream)?;
    let mut players = server.players.lock().unwrap();
    let player = Player {
        role,
        username,
        public_key,
    };
    players.insert(
        server.id,
        (player.clone(), server.others_serv_msg_buf.clone()),
    );
    println!("Created player {}", server.id);
    Ok(player)
}

fn receive_user_role(
    server: &Server,
    stream: &mut UnixStream,
) -> Result<UserRole, RecvPositionError> {
    let mut chosen_role: UserRole;
    loop {
        //send available roles
        let available_roles = compute_available_roles_to_str(server);
        println!("Computed available_roles as str: {}", available_roles);
        clichess::write_to_stream(stream, available_roles.clone()).unwrap();
        //receive chosen role
        let chosen_role_str = server
            .client_message_recv
            .recv()
            .expect("Player closed connection.");
        if chosen_role_str.trim() == EXIT_MSG {
            return Err(RecvPositionError::UserCanceledError);
        }
        println!("Client id {} has chosen {}", server.id, chosen_role_str);
        let chosen_role_parse = clichess::parse_to_role(&chosen_role_str);

        if chosen_role_parse.is_err() {
            return Err(RecvPositionError::ParsePositionError);
        }
        chosen_role = chosen_role_parse.unwrap();
        //check if role is still available after the choice
        {
            let mut available_roles_after_choice =
                vec![UserRole::White, UserRole::Black, UserRole::Spectator];
            let players = server.players.lock().unwrap();
            for (_, (player, _)) in players.iter() {
                match available_roles_after_choice
                    .iter()
                    .position(|r| *r == player.role)
                {
                    Some(index) => available_roles_after_choice.remove(index),
                    None => continue,
                };
            }
            if available_roles_after_choice.contains(&chosen_role) {
                println!("OK");
                clichess::write_to_stream(stream, String::from("OK")).unwrap();
                server
                    .client_message_recv
                    .recv()
                    .expect("Player closed connection.");
                break;
            } else {
                println!("KO");
                clichess::write_to_stream(stream, String::from("KO")).unwrap();
                server
                    .client_message_recv
                    .recv()
                    .expect("Player closed connection.");
                continue;
            }
        }
    }
    Ok(chosen_role)
}

fn compute_available_roles_to_str(server: &Server) -> String {
    let mut available_roles = vec![UserRole::White, UserRole::Black, UserRole::Spectator];
    let players = server.players.lock().unwrap();
    for (_, (player, _)) in players.iter() {
        match available_roles.iter().position(|r| *r == player.role) {
            Some(index) => available_roles.remove(index),
            None => continue,
        };
    }
    let available_roles_str: Vec<String> = available_roles
        .iter()
        .map(|r| clichess::role_to_str(r))
        .collect();
    available_roles_str.join(",")
}

fn player_disconnected(server: &Server) {
    let mut players = server.players.lock().unwrap();
    let player_and_receiving_buf = players.get(&server.id);
    if player_and_receiving_buf.is_some() {
        let (player, _) = player_and_receiving_buf.expect("is some");
        println!("Player {} disconnected.", player.username);
    }
    players.remove(&server.id);
}

fn setup_client_message_recv(stream: &UnixStream) -> io::Result<Receiver<String>> {
    let (sender, receiver) = channel();
    let thread_stream = stream.try_clone()?;

    thread::spawn(move || {
        loop {
            //wait for client message
            let buffer =
                clichess::read_line_from_stream(&thread_stream).expect("Error message from server");
            let send_result = sender.send(buffer);
            // stop the thread during a disconnection
            if send_result.is_err() {
                break;
            }
        }
    });
    Ok(receiver)
}
